﻿module LambdaCalculus

(* Definition *)

type Term =   
    Variable     of string
   | Lambda      of string * Term 
   | Application of Term * Term
   
(* Auswertung *) 

let rec freeVars term = 
    match term with
    | Variable x       -> Set.singleton x
    | Application(f,x) -> Set.union (freeVars f) (freeVars x)
    | Lambda(x,b)      -> Set.difference (freeVars b) (Set.singleton x)

let freshVar = 
    let possible = seq {
        for chr in 'a'..'z' do
            yield chr.ToString()
        for n in Seq.initInfinite(fun i -> i + 1) do
            yield sprintf "a%i" n
    }

    fun notIn -> Seq.find (fun elem -> not (Set.contains elem notIn)) possible
    
let rec sub term x expn =
    match term with
    | Variable y when y = x -> expn
    | Variable y -> term
    | Application(f, a) -> Application(sub f x expn, sub a x expn)
    | Lambda(y, body) when y = x -> term 
    | Lambda(y, body) ->
        let free = freeVars expn
        let doesCapture = free |> Set.contains y
        if doesCapture then
            let newY = freshVar (Set.union free (freeVars term))
            Lambda(newY, sub (sub body y (Variable newY)) x expn)
        else
            Lambda(y, sub body x expn)

let rec beta term =
    match term with
    | Application(Lambda(x, body), expn) -> (sub body x expn, true)
    | Application(a, b) -> 
        let (a', c1) = beta a
        let (b', c2) = beta b
        (Application(a',b'), c1 || c2)
    // Eta reduction
    | Lambda(x, Application(body, Variable(y))) when x = y -> (body, true)
    | _ -> (term, false)

let rec eval term = 
    let (newTerm, hasChanged) = beta term 
    if (hasChanged) then
        eval newTerm
    else
        newTerm

let rec prettyPrint term = 
    let parenL = function
        | Lambda(_) as l -> sprintf "(%s)" (prettyPrint l)
        | l -> prettyPrint l
    let parenR = function
        | Variable x -> x
        | r -> sprintf "(%s)" (prettyPrint r)
    match term with
    | Variable x -> x
    | Lambda(x, body) -> sprintf "\\%s . %s" x (prettyPrint body)
    | Application(l, r) -> sprintf "%s %s" (parenL l) (parenR r)

let escape (varname : string) =
    let bytes = System.Text.Encoding.UTF8.GetBytes(varname)
    let parts =  [| 
            for byte in bytes do
                yield System.Convert.ToString(byte, 16)
        |]   
    "_" + System.String.Join("", parts) 

let rec makeJS term = 
    match term with
    | Variable(x) -> escape x
    | Lambda(x, body) -> sprintf "function(%s) { return %s; }" (escape x) (makeJS body)
    | Application(a, b) -> sprintf "(%s)(%s)" (makeJS a) (makeJS b)

let ydef term = "(\ Y . (" + term + "))(\ f . (\ x . f (x x)) (\ x . f (x x)))"